package be.jeffcheasey88.peeratcode.parser.state;

import java.util.ArrayList;
import java.util.List;
import java.util.function.BiFunction;
import java.util.function.Function;

import be.jeffcheasey88.peeratcode.parser.Bag;
import be.jeffcheasey88.peeratcode.parser.TokenValidator;

public class StateTree<E>{
	
	private List<StateTree<E>> childs;
	
	Function<TokenValidator, Boolean> checker;
	private BuilderStateTree<E, ?> builder;
	
	public StateTree(){
		this.childs = new ArrayList<>();
	}
	
	public void seed(TokenValidator validator, E container){
		while(validator.hasNext()){
			BuilderStateTree<E, ?> build = internalSeed(validator, container);
			if(build == null) break;
			build.build(validator, container);
		}
	}
	
	BuilderStateTree<E, ?> internalSeed(TokenValidator validator, E element){
		for(StateTree<E> child : this.childs){
			TokenValidator branch = validator.branch();
			if(child.checker == null){
				BuilderStateTree<E, ?> builded = child.internalSeed(branch, element);
				if(builded != null){
					validator.merge(branch);
					return builded;
				}
				continue;
			}
			if(child.checker.apply(branch)){
				BuilderStateTree<E, ?> builded = child.internalSeed(branch, element);
				if(builded != null){
					validator.merge(branch);
					return builded;
				}
			}
		}
		if(this.builder != null) return this.builder;
		return null;
	}

	public <T extends StateTree<E>> StateTree<E> then(StateTree<E> child){
		childs.add(child);
		return child;
	}
	
	public <T extends StateTree<E>> StateTree<E> then(Function<TokenValidator, Boolean> checker){
		StateTree<E> state = new StateTree<>();
		state.checker = checker;
		this.childs.add(state);
		return state;
	}
	
	public <T extends StateTree<E>> StateTree<E> loop(){
		this.childs.add(this);
		return this;
	}
	
	public <B> BuilderStateTree<E, B> end(){
		return end((a,b) -> null);
	}
	
	public <B> BuilderStateTree<E, B> end(BiFunction<E, Bag, B> builder){
		BuilderStateTree<E, B> builderState = new BuilderStateTree<>(builder);
		this.builder = builderState;
		return builderState;
	}
	
}
