package be.jeffcheasey88.peeratcode.parser.java;

import static org.junit.jupiter.api.Assertions.assertEquals;

import java.util.function.BiConsumer;

import org.junit.jupiter.api.Test;

import be.jeffcheasey88.peeratcode.parser.Bag;
import be.jeffcheasey88.peeratcode.parser.Parser;
import be.jeffcheasey88.peeratcode.parser.Token;
import be.jeffcheasey88.peeratcode.parser.TokenType;
import be.jeffcheasey88.peeratcode.parser.TokenValidator;
import be.jeffcheasey88.peeratcode.parser.state.RedirectStateTree;
import be.jeffcheasey88.peeratcode.parser.state.StateTree;

public class TypeTests{
	
	private static StateTree<JavaElement> type;
	private static StateTree<JavaElement> gen;
	
	static{
		BiConsumer<Bag, Token> concat = (bag, token) -> {
			Token value = bag.get();
			if(value == null) bag.set(token);
			else bag.set(value.concat(token));
		};
		
		type = new StateTree<JavaElement>();
		StateTree<JavaElement> type_ = type.then((validator) -> validator.validate((token) -> token.getType().equals(TokenType.NAME),concat));
		type_.then((validator) -> validator.validate((token) -> token.getValue().equals("."), concat))
			.then(type_);
		gen = new StateTree<>();
		type_.then(gen);
		StateTree<JavaElement> type_begin = gen.then((validator) -> validator.validate((token) -> token.getValue().equals("<"), concat));
		StateTree<JavaElement> type_generic = type_begin.then(new RedirectStateTree<>(type, (global, local) -> global.set(global.<Token>get().concat(local.get()))));
		StateTree<JavaElement> type_generic_end = type_generic.then((validator) -> validator.validate((token) -> token.getValue().equals(">"), concat));
		StateTree<JavaElement> type_generic_split = type_generic.then((validator) -> validator.validate((token) -> token.getValue().equals(","), concat));
		type_generic_split.then(type_generic);
		
		type_begin.then(type_generic_end);
		
		StateTree<JavaElement> type_generic_unknow = type_begin.then((validator) -> validator.validate((token) -> token.getValue().equals("?"), concat));
		type_generic_unknow.then(type_generic_end);
		type_generic_split.then(type_generic_unknow);
		type_generic_unknow.then(type_generic_split);
		StateTree<JavaElement> type_generic_unknow_extends = type_generic_unknow.then((validator) -> validator.validate((token) -> token.getValue().equals("extends"), concat))
			.then(new RedirectStateTree<>(type, (global, local) -> global.set(global.<Token>get().concat(local.get()))));
		type_generic_unknow_extends.then(type_generic_end);
		type_generic_unknow_extends.then(type_generic_split);
		StateTree<JavaElement> type_generic_unknow_super = type_generic_unknow.then((validator) -> validator.validate((token) -> token.getValue().equals("super"), concat))
		.then(new RedirectStateTree<>(type, (global, local) -> global.set(global.<Token>get().concat(local.get()))));
		type_generic_unknow_super.then(type_generic_end);
		type_generic_unknow_super.then(type_generic_split);
		
		StateTree<JavaElement> type_generic_named_extends = type_generic.then((validator) -> validator.validate((token) -> token.getValue().equals("extends"), concat))
				.then(new RedirectStateTree<>(type, (global, local) -> global.set(global.<Token>get().concat(local.get()))));
		type_generic_named_extends.then(type_generic_end);
		type_generic_named_extends.then(type_generic_split);
		StateTree<JavaElement> type_generic_named_super = type_generic.then((validator) -> validator.validate((token) -> token.getValue().equals("super"), concat))
				.then(new RedirectStateTree<>(type, (global, local) -> global.set(global.<Token>get().concat(local.get()))));
		type_generic_named_super.then(type_generic_end);
		type_generic_named_super.then(type_generic_split);
		
		type_.end();
		type_generic_end.end();
	}
	
	public static StateTree<JavaElement> get(){
		return type;
	}
	
	public static StateTree<JavaElement> getGeneric(){
		return gen;
	}
	
	
	private static Parser<JavaElement> parser = new Parser<JavaElement>(){
		{
			setTokenizer(ModifierTests.TOKENIZER);
			
			setStateTree(get());
		}
	};
	
	JavaElement testCase(String value) throws Exception{
		TokenValidator.TOKENS = 0;
		TokenValidator.MAX_VALIDATE = 0;
		
		JavaElement result = new JavaElement();
		
		parser.parse(value, result);
		
		assertEquals(TokenValidator.TOKENS, TokenValidator.MAX_VALIDATE);
		
		return result;
	}
	
	@Test
	void baseType() throws Exception{
		JavaElement element = testCase("TokenType");
		
	}
	
	@Test
	void baseGenericType() throws Exception{
		JavaElement element = testCase("TokenType<E>");
		
	}
	
	@Test
	void genericType() throws Exception{
		JavaElement element = testCase("TokenType<Map<String,E extends Test>>");
		
	}

}
