package be.jeffcheasey88.peeratcode.parser.java;

import static org.junit.jupiter.api.Assertions.assertEquals;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.BiFunction;

import org.junit.jupiter.api.Test;

import be.jeffcheasey88.peeratcode.parser.Bag;
import be.jeffcheasey88.peeratcode.parser.Parser;
import be.jeffcheasey88.peeratcode.parser.Token;
import be.jeffcheasey88.peeratcode.parser.TokenType;
import be.jeffcheasey88.peeratcode.parser.TokenValidator;
import be.jeffcheasey88.peeratcode.parser.java.Variable.VariableContainer;
import be.jeffcheasey88.peeratcode.parser.state.InitialStateTree;
import be.jeffcheasey88.peeratcode.parser.state.RedirectStateTree;
import be.jeffcheasey88.peeratcode.parser.state.StateTree;

public class VariableTests{
	
	public static StateTree<JavaElement> get(){
		BiFunction<JavaElement, Bag, JavaElement> variable_builder = (parent, bag) -> {
			Integer mod = bag.get("mod");
			Token type = bag.get("type");
			Map<Token, Value> map = bag.get("vars");
			for(Entry<Token, Value> vars : map.entrySet()){
				Variable result = new Variable(mod == null ? 0 : mod, type, vars.getKey(), false, vars.getValue());
				bag.set(result);
				if(parent instanceof VariableContainer) ((VariableContainer)parent).addVariable(result);
			}
			return null;
		};
		
		InitialStateTree<JavaElement> variable = new InitialStateTree<>();
		variable.multiple(AnnotationTests.get());
		StateTree<JavaElement> variable_mod = variable.then(new RedirectStateTree<>(ModifierTests.get(), (global, local) -> global.set("mod", local.get())));
		StateTree<JavaElement> variable_type = variable.then(new RedirectStateTree<>(TypeTests.get(), (global, local) -> global.set("type", local.get())));
		variable_mod.then(variable_type);
		StateTree<JavaElement> variable_name = variable_type.then((validator) -> validator.validate(
				(token) -> token.getType().equals(TokenType.NAME),
				(bag, token) -> {
					Map<Token, Value> map = bag.get("vars");
					if(map == null){
						map = new LinkedHashMap<>();
						bag.set("vars", map);
					}
					bag.set("last", token);
					map.put(token, null);
				}));
		variable_name.then((validator) -> validator.validate((token) -> token.getValue().equals(";")))
			.end(variable_builder);
		StateTree<JavaElement> variable_split = variable_name.then((validator) -> validator.validate((token) -> token.getValue().equals(",")));
		variable_split.then(variable_name);
		StateTree<JavaElement> variable_value = variable_name.then((validator) -> validator.validate((token) -> token.getValue().equals("=")))
			.then(new RedirectStateTree<>(ValueTests.get(), (global, local) -> {
				global.<Map<Token, Value>>get("vars").put(global.get("last"), local.get());
			}));
		variable_value.then(variable_split);
		variable_value.then((validator) -> validator.validate((token) -> token.getValue().equals(";")))
			.end(variable_builder);
		variable_value.end(variable_builder);
		
		return variable;
	}
	
	private static Parser<JavaElement> parser = new Parser<JavaElement>(){
		{
			setTokenizer(ModifierTests.TOKENIZER);
			
			setStateTree(get());
		}
	};
	
	JavaElement testCase(String value) throws Exception{
		TokenValidator.TOKENS = 0;
		TokenValidator.MAX_VALIDATE = 0;
		
		JavaElement result = new JavaElement();
		
		parser.parse(value, result);
		
		assertEquals(TokenValidator.TOKENS, TokenValidator.MAX_VALIDATE);
		
		return result;
	}

	@Test
	void hashMap() throws Exception{
		JavaElement result = testCase("private static final Map<Map<List<String>>> map = new HashMap<>();");
	}
}
